/*	Direct3DDevice OpenGL driver
 *
 * Copyright 2001 TransGaming Technologies Inc.
 */

/* This is not a self-contained driver - in the same way that OpenGL requires
 * an OS-specific interface like GLX or WGL to operate, this GL driver is
 * supposed to be bolted onto another Direct3DDevice driver (e.g. the HAL driver)
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>

#define USE_NEW_HAL /* temporary */

#include "debugtools.h"
#include "ddrawi.h"
#include "d3dhal.h"
#include "d3d_private.h"
#include "ddraw/main.h"
#include "ddraw/user.h"
#include "ddraw/hal.h"
#include "ddraw/gl.h"
#include "d3ddevice/main.h"
#include "d3ddevice/gl.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

static HRESULT WINAPI GL_alloc_state_block(LPVOID ctx, LPDWORD lpdwBlock);
static HRESULT WINAPI GL_free_state_block(LPVOID ctx, DWORD dwBlock);
static void WINAPI GL_demolish_texture(LPVOID ctx, DWORD dwTexture);
static HRESULT WINAPI GL_start_dp2(LPVOID ctx, DWORD dwFlags,
				   DWORD dwVertexType, LPVOID lpVertices,
				   DWORD dwVertexOffset, DWORD dwVertexLength,
				   DWORD dwReqVertexBufSize, DWORD dwReqCommandBufSize);
static HRESULT WINAPI GL_start_dp2_sd(LPVOID ctx, DWORD dwFlags,
				      DWORD dwVertexType, LPVOID lpVertices,
				      DWORD dwVertexOffset, DWORD dwVertexLength,
				      DWORD dwReqVertexBufSize, DWORD dwReqCommandBufSize,
				      LPD3DDRAWPRIMITIVESTRIDEDDATA lpStridedData);
static HRESULT WINAPI GL_emit_dp2(LPVOID ctx, LPD3DHAL_DP2COMMAND cmd,
				  LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2);
static HRESULT WINAPI GL_flush_dp2(LPVOID ctx);

static HRESULT GL_replay_cmds(IDirect3DDeviceImpl *This, struct GL_D3D_saved_cmd **cmd, DWORD dwFlags);
#define REPLAY_DEL	1
#define REPLAY_EXEC	2
#define REPLAY_APPLY	4

LPVOID WINAPI GL_Direct3DDevice_Construct(IDirect3DDeviceImpl* This)
{
    GL_D3D_PRIV_VAR(priv, This);
    LPDDRAWI_DIRECTDRAW_GBL dd_gbl = This->ddraw_owner->local.lpGbl;

    TRACE("(%p)\n",This);

    priv->gl.gl_data = (DD_GL_data*)dd_gbl->dwReserved3;
    priv->gl.lpCaps = (LPD3DHAL_D3DEXTENDEDCAPS)dd_gbl->lpD3DExtendedCaps;
    priv->gl.glFlags = GLF_NEW_TRANSFORM | GLF_NEW_LIGHTING;
    priv->gl.material_flags = GLMF_ALL_MATERIAL_DIRTY;
    priv->gl.texture_flags = GLTF_DEFAULT_TEXCOORD_TRANSFORM;

    priv->gl.color_array_size = 8192;
    priv->gl.color_array = VirtualAlloc(NULL, priv->gl.color_array_size,
					MEM_COMMIT | MEM_RESERVE,
					PAGE_READWRITE);
    priv->gl.vertex_array_size = 8192;
    priv->gl.vertex_array = VirtualAlloc(NULL, priv->gl.vertex_array_size,
					 MEM_COMMIT | MEM_RESERVE,
					 PAGE_READWRITE);

    /* record "init" command */
    priv->gl.cmds = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(struct GL_D3D_saved_cmd));
    priv->gl.cmds->cmd.bCommand = 0;

    This->alloc_state_block = GL_alloc_state_block;
    This->free_state_block = GL_free_state_block;
    This->demolish_texture = GL_demolish_texture;
    This->start_dp2 = GL_start_dp2;
    This->start_dp2_sd = GL_start_dp2_sd;
    This->emit_dp2 = GL_emit_dp2;
    This->flush_dp2 = GL_flush_dp2;

    This->validate_device = GL_Direct3DDevice_validate_device;

    return This;
}

void WINAPI GL_Direct3DDevice_Destruct(LPVOID ctx)
{
    IDirect3DDeviceImpl* This = (IDirect3DDeviceImpl*)ctx;
    GL_D3D_PRIV_VAR(priv, This);
    VirtualFree(priv->gl.color_array, 0, MEM_RELEASE);
    VirtualFree(priv->gl.vertex_array, 0, MEM_RELEASE);
    while (priv->gl.cmds) {
	struct GL_D3D_saved_cmd *ncmd = priv->gl.cmds;
	priv->gl.cmds = ncmd->next;
	if (ncmd->data1) HeapFree(GetProcessHeap(),0,ncmd->data1);
	if (ncmd->data2) HeapFree(GetProcessHeap(),0,ncmd->data2);
	HeapFree(GetProcessHeap(),0,ncmd);
    }
}

/* Talk to the driver using the HAL GetDriverInfo interface. */
static void D3D_GL_query_driver(IDirect3DDeviceImpl* This)
{
    GL_D3D_PRIV_VAR(priv, This);
    HAL_DirectDraw_GetDriverInfo(This->ddraw_owner, &GUID_WineGLD3DCallbacks,
				 &priv->gl.callbacks,
				 sizeof(priv->gl.callbacks));
}

static const float zero4f[4] = { 0, 0, 0, 0 };
static const float one4f[4] = { 1, 1, 1, 1 };

/* This routine should only initialize anything non-render-state-specific
 * (the default render states are set separately) */
HRESULT D3D_GL_init(IDirect3DDeviceImpl* This, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    static const GLfloat sgenparams[] = {1.0, 0.0, 0.0, 0.0};
    static const GLfloat tgenparams[] = {0.0, 1.0, 0.0, 0.0};
    static const GLfloat rgenparams[] = {0.0, 0.0, 1.0, 0.0};
    GL_D3D_PRIV_VAR(priv, This);
    unsigned u;

    D3D_GL_query_driver(This);

    TRACE("initializing GL context\n");

#ifndef HAVE_OSMESA
    glFrontFace(GL_CW);
#else
    /* OSMESA y-swap */
    glFrontFace(GL_CCW);
#endif

    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, one4f);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, zero4f);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, zero4f);
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, zero4f);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, zero4f);
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
    priv->gl.current_color_material = GL_AMBIENT_AND_DIFFUSE;

    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

    glDisable(GL_DITHER);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    for (u=0; u < priv->gl.lpCaps->wMaxTextureBlendStages; u++) {
	GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	glTexGenfv(GL_S, GL_EYE_PLANE, sgenparams);
	glTexGenfv(GL_T, GL_EYE_PLANE, tgenparams);
	glTexGenfv(GL_R, GL_EYE_PLANE, rgenparams);
    }

    glPopMatrix();

    /* if texture_env_combine is available, enable it */
    if (HAVE_GL_EXT(priv, ARB_texture_env_combine) || HAVE_GL_EXT(priv, EXT_texture_env_combine)) {
	for (u=0; u<priv->gl.lpCaps->wMaxTextureBlendStages; u++) {
	    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
	}
    }

    return D3D_OK;
}

static HRESULT WINAPI GL_alloc_state_block(LPVOID ctx, LPDWORD lpdwBlock)
{
    struct GL_D3D_state_block *block = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
						 sizeof(struct GL_D3D_state_block));
    TRACE("creating state block %p\n", block);
    /* won't bother to allocate a handle yet, just cast the pointer... */
    *lpdwBlock = (DWORD)block;
    return D3D_OK;
}

static HRESULT WINAPI GL_free_state_block(LPVOID ctx, DWORD dwBlock)
{
    /* don't do anything here, let D3DHAL_STATESETDELETE take care of it */
    return D3D_OK;
}

static BOOL GL_update_cmd(struct GL_D3D_saved_cmd *ocmd,
			  LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    if (ocmd->cmd.bCommand != cmd->bCommand) return FALSE;
    switch (ocmd->cmd.bCommand) {
    case D3DDP2OP_RENDERSTATE:
	{
	    LPD3DHAL_DP2RENDERSTATE stat1 = ocmd->data1;
	    LPD3DHAL_DP2RENDERSTATE stat2 = data1;
	    if (stat1->RenderState == stat2->RenderState) {
		stat1->u.dwState = stat2->u.dwState;
		return TRUE;
	    }
	}
	break;
    case D3DDP2OP_TEXTURESTAGESTATE:
	{
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat1 = ocmd->data1;
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat2 = data1;
	    if (stat1->wStage != stat2->wStage) break;
	    if (stat1->TSState == stat2->TSState) {
		stat1->dwValue = stat2->dwValue;
		return TRUE;
	    }
	}
	break;
    case D3DDP2OP_VIEWPORTINFO:
	{
	    LPD3DHAL_DP2VIEWPORTINFO stat1 = ocmd->data1;
	    LPD3DHAL_DP2VIEWPORTINFO stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    case D3DDP2OP_WINFO:
	{
	    LPD3DHAL_DP2WINFO stat1 = ocmd->data1;
	    LPD3DHAL_DP2WINFO stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    /* FIXME: SETPALETTE, UPDATEPALETTE */
    case D3DDP2OP_ZRANGE:
	{
	    LPD3DHAL_DP2ZRANGE stat1 = ocmd->data1;
	    LPD3DHAL_DP2ZRANGE stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    case D3DDP2OP_SETMATERIAL:
	{
	    LPD3DHAL_DP2SETMATERIAL stat1 = ocmd->data1;
	    LPD3DHAL_DP2SETMATERIAL stat2 = data1;
	    *stat1 = *stat2;
	    return TRUE;
	}
	break;
    case D3DDP2OP_SETLIGHT:
	{
	    LPD3DHAL_DP2SETLIGHT stat1 = ocmd->data1;
	    LPD3DHAL_DP2SETLIGHT stat2 = data1;
	    if (stat1->dwIndex != stat2->dwIndex) return FALSE;
	    if (stat1->dwDataType < D3DHAL_SETLIGHT_DATA &&
		stat2->dwDataType < D3DHAL_SETLIGHT_DATA) {
		stat1->dwDataType = stat2->dwDataType;
		return TRUE;
	    }
	    if (stat1->dwDataType == D3DHAL_SETLIGHT_DATA &&
		stat2->dwDataType == D3DHAL_SETLIGHT_DATA) {
		LPD3DLIGHT7 l1 = ocmd->data1;
		LPD3DLIGHT7 l2 = data2;
		*l1 = *l2;
		return TRUE;
	    }
	}
	break;
    case D3DDP2OP_CREATELIGHT:
	{
	    LPD3DHAL_DP2CREATELIGHT stat1 = ocmd->data1;
	    LPD3DHAL_DP2CREATELIGHT stat2 = data1;
	    if (stat1->dwIndex < stat2->dwIndex)
		stat1->dwIndex = stat2->dwIndex;
	    return TRUE;
	}
	break;
    case D3DDP2OP_SETTRANSFORM:
	{
	    LPD3DHAL_DP2SETTRANSFORM stat1 = ocmd->data1;
	    LPD3DHAL_DP2SETTRANSFORM stat2 = data1;
	    if (stat1->xfrmType == stat2->xfrmType) {
		stat1->matrix = stat2->matrix;
		return TRUE;
	    }
	}
	break;
    /* FIXME: SETPRIORITY, SETTEXLOD, SETCLIPPLANE */
    default:
	break;
    }
    return FALSE;
}

static void GL_capture_cmd(IDirect3DDeviceImpl *This,
			   struct GL_D3D_saved_cmd *ocmd)
{
    /* FIXME */
}

static void GL_apply_cmd(IDirect3DDeviceImpl *This,
			 struct GL_D3D_saved_cmd *ocmd)
{
    switch (ocmd->cmd.bCommand) {
    case D3DDP2OP_RENDERSTATE:
	{
	    LPD3DHAL_DP2RENDERSTATE stat = ocmd->data1;
	    This->d3dp.renderstate[stat->RenderState] = stat->u.dwState;
	}
	break;
    case D3DDP2OP_TEXTURESTAGESTATE:
	{
	    LPD3DHAL_DP2TEXTURESTAGESTATE stat = ocmd->data1;
	    This->d3dp.texstagestate[stat->wStage][stat->TSState] = stat->dwValue;
	}
	break;
    case D3DDP2OP_VIEWPORTINFO:
	{
	    LPD3DHAL_DP2VIEWPORTINFO stat = ocmd->data1;
	    This->d3dp.viewport.dwX = stat->dwX;
	    This->d3dp.viewport.dwY = stat->dwY;
	    This->d3dp.viewport.dwWidth = stat->dwWidth;
	    This->d3dp.viewport.dwHeight = stat->dwHeight;
	}
	break;
    case D3DDP2OP_WINFO:
	/* ? */
	break;
    case D3DDP2OP_ZRANGE:
	{
	    LPD3DHAL_DP2ZRANGE stat = ocmd->data1;
	    This->d3dp.viewport.dvMinZ = stat->dvMinZ;
	    This->d3dp.viewport.dvMaxZ = stat->dvMaxZ;
	}
	break;
    case D3DDP2OP_SETMATERIAL:
	{
	    LPD3DHAL_DP2SETMATERIAL stat = ocmd->data1;
	    This->d3dp.material = *stat;
	}
	break;
    case D3DDP2OP_SETLIGHT:
	{
	    LPD3DHAL_DP2SETLIGHT stat = ocmd->data1;
	    switch (stat->dwDataType) {
	    case D3DHAL_SETLIGHT_ENABLE:
		This->d3dp.light[stat->dwIndex].flags |= D3DLIGHT_ACTIVE;
		break;
	    case D3DHAL_SETLIGHT_DISABLE:
		This->d3dp.light[stat->dwIndex].flags &= ~D3DLIGHT_ACTIVE;
		break;
	    case D3DHAL_SETLIGHT_DATA:
		This->d3dp.light[stat->dwIndex].l = *(LPD3DLIGHT7)ocmd->data2;
		break;
	    }
	}
	break;
    case D3DDP2OP_SETTRANSFORM:
	{
	    LPD3DHAL_DP2SETTRANSFORM stat = ocmd->data1;
	    This->d3dp.xfrm[stat->xfrmType] = stat->matrix;
	}
	break;
    /* FIXME: SETCLIPPLANE */
    }
}

static struct GL_D3D_saved_cmd*
GL_duplicate_cmd(IDirect3DDeviceImpl *This, LPD3DHAL_DP2COMMAND cmd,
		 LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2)
{
    struct GL_D3D_saved_cmd *ncmd;

    ncmd = HeapAlloc(GetProcessHeap(),0,sizeof(struct GL_D3D_saved_cmd));
    if (!ncmd) return ncmd;
    ncmd->next = NULL;
    ncmd->cmd = *cmd;
    ncmd->data1 = data1 ? HeapAlloc(GetProcessHeap(),0,dwSize1) : NULL;
    if (ncmd->data1) memcpy(ncmd->data1, data1, dwSize1);
    ncmd->data2 = data2 ? HeapAlloc(GetProcessHeap(),0,dwSize2) : NULL;
    if (ncmd->data2) memcpy(ncmd->data2, data2, dwSize2);
    return ncmd;
}

static void
GL_record_cmd(IDirect3DDeviceImpl *This, struct GL_D3D_saved_cmd **cmds,
	      LPD3DHAL_DP2COMMAND cmd, LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2)
{
    struct GL_D3D_saved_cmd *ncmd = *cmds, *lcmd = NULL;

    while (ncmd) {
	if (GL_update_cmd(ncmd, cmd, data1, data2))
	    return;
	lcmd = ncmd;
	ncmd = ncmd->next;
    }
    ncmd = GL_duplicate_cmd(This, cmd, data1, dwSize1, data2, dwSize2);
    if (lcmd) lcmd->next = ncmd;
    else *cmds = ncmd;
}

static void GL_delete_cmd(IDirect3DDeviceImpl *This, struct GL_D3D_saved_cmd **cmd)
{
    struct GL_D3D_saved_cmd *ncmd = *cmd;
    *cmd = ncmd->next;
    if (ncmd->data1) HeapFree(GetProcessHeap(),0,ncmd->data1);
    if (ncmd->data2) HeapFree(GetProcessHeap(),0,ncmd->data2);
    HeapFree(GetProcessHeap(),0,ncmd);
}

static void WINAPI GL_demolish_texture(LPVOID ctx, DWORD dwTexture)
{
    IDirect3DDeviceImpl* This = (IDirect3DDeviceImpl*)ctx;
    GL_D3D_PRIV_VAR(priv, This);
    struct GL_D3D_saved_cmd **cmd = &priv->gl.cmds;

    while (*cmd) {
	BOOL kill_it = FALSE;
	switch ((*cmd)->cmd.bCommand) {
	case D3DDP2OP_TEXBLT:
	    {
		LPD3DHAL_DP2TEXBLT dat = (*cmd)->data1;
		if (dat->dwDDDestSurface == dwTexture ||
		    dat->dwDDSrcSurface == dwTexture) kill_it = TRUE;
	    }
	    break;
	default:
	    break;
	}
	if (kill_it)
	    GL_delete_cmd(This, cmd);
	else
	    cmd = &(*cmd)->next;
    }
}

static HRESULT D3D_GL_state_set(IDirect3DDeviceImpl *This, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2STATESET dat = (LPD3DHAL_DP2STATESET)data1;
    struct GL_D3D_state_block *block = (struct GL_D3D_state_block *)dat->dwParam;
    GL_D3D_PRIV_VAR(priv, This);

    switch (dat->dwOperation) {
    case D3DHAL_STATESETBEGIN:
	TRACE("recording to state block %p\n", block);
	priv->gl.block = block;
	break;
    case D3DHAL_STATESETEND:
	TRACE("state block %p complete\n", block);
	priv->gl.block = NULL;
	break;
    case D3DHAL_STATESETDELETE:
	TRACE("deleting state block %p\n", block);
	if (block->list)
	    glDeleteLists(block->list, 1);
	while (block->cmds) GL_delete_cmd(This, &block->cmds);
	HeapFree(GetProcessHeap(), 0, block);
	break;
    case D3DHAL_STATESETEXECUTE:
	TRACE("executing state block %p\n", block);
	if (!block->list) {
	    block->list = glGenLists(1);
	    glNewList(block->list, GL_COMPILE_AND_EXECUTE);
	    /* FIXME: lights might not work reliably this way? */
	    GL_replay_cmds(This, &block->cmds, REPLAY_EXEC|REPLAY_APPLY);
	    glEndList();
	} else {
	    glCallList(block->list);
	    GL_replay_cmds(This, &block->cmds, REPLAY_APPLY);
	}
	break;
    case D3DHAL_STATESETCAPTURE:
	FIXME("STATESETCAPTURE not implemented\n");
	break;
    }
    return D3D_OK;
}

typedef HRESULT (*GL_dp2_handler)(IDirect3DDeviceImpl *This, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2);

#define PRIM 1
#define STAT 2
#define OTHR 3

static const struct {
    GL_dp2_handler func;
    int type;
    const char*name;
} GL_dp2_func[] = {
    {D3D_GL_init,		OTHR, "INIT"},
    {D3D_GL_draw,		PRIM, "POINTS"},
    {NULL,			PRIM, "INDEXEDLINELIST"},
    {NULL,			PRIM, "INDEXEDTRIANGLELIST"},
    {NULL}, {NULL}, {NULL}, {NULL},
    {D3D_GL_render_state, 	STAT, "RENDERSTATE"},
    {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL},
    {D3D_GL_draw,		PRIM, "LINELIST"},
    {D3D_GL_draw,		PRIM, "LINESTRIP"},
    {D3D_GL_draw,		PRIM, "INDEXEDLINESTRIP"},
    {D3D_GL_draw,		PRIM, "TRIANGLELIST"},
    {D3D_GL_draw,		PRIM, "TRIANGLESTRIP"},
    {D3D_GL_draw,		PRIM, "INDEXEDTRIANGLESTRIP"},
    {D3D_GL_draw,		PRIM, "TRIANGLEFAN"},
    {D3D_GL_draw,		PRIM, "INDEXEDTRIANGLEFAN"},
    {NULL,			PRIM, "TRIANGLEFAN_IMM"},
    {NULL,			PRIM, "LINELIST_IMM"},
    {D3D_GL_texture_stage_state,STAT, "TEXTURESTAGESTATE"},
    {D3D_GL_draw,		PRIM, "INDEXEDTRIANGLELIST2"},
    {D3D_GL_draw,		PRIM, "INDEXEDLINELIST2"},
    {D3D_GL_viewport_info,	STAT, "VIEWPORTINFO"},
    {NULL,			STAT, "WINFO"},
    {NULL,			OTHR, "SETPALETTE"},
    {NULL,			OTHR, "UPDATEPALETTE"},
    {D3D_GL_z_range,		STAT, "ZRANGE"},
    {D3D_GL_set_material,	STAT, "SETMATERIAL"},
    {D3D_GL_set_light,		STAT, "SETLIGHT"},
    {D3D_GL_create_light,	OTHR, "CREATELIGHT"},
    {D3D_GL_set_transform,	STAT, "SETTRANSFORM"},
    {NULL,			OTHR, "EXT"},
    {D3D_GL_texblt,		OTHR, "TEXBLT"},
    {D3D_GL_state_set,		OTHR, "STATESET"},
    {NULL,			OTHR, "SETPRIORITY"},
    {NULL,			OTHR, "SETRENDERTARGET"},
    {D3D_GL_clear,		OTHR, "CLEAR"},
    {NULL,			OTHR, "SETTEXLOD"},
    {NULL,			STAT, "SETCLIPPLANE"}
};

static HRESULT WINAPI GL_start_dp2_sd(LPVOID ctx, DWORD dwFlags,
				      DWORD dwVertexType, LPVOID lpVertices,
				      DWORD dwVertexOffset, DWORD dwVertexLength,
				      DWORD dwReqVertexBufSize, DWORD dwReqCommandBufSize,
				      LPD3DDRAWPRIMITIVESTRIDEDDATA lpStridedData)
{
    IDirect3DDeviceImpl* This = (IDirect3DDeviceImpl*)ctx;
    GL_D3D_PRIV_VAR(priv, This);

    priv->gl.glFlags &= ~(GLF_STRIDED_DATA|GLF_VERTEX_BUFFER);
    priv->gl.dp2_VertexBuffer = NULL;
    priv->gl.dp2_lpVertices = NULL;

    if (lpStridedData)
    {
	priv->gl.sd = *lpStridedData;
	priv->gl.glFlags |= GLF_STRIDED_DATA;
    }
    else if (!(dwFlags & D3DHALDP2_USERMEMVERTICES))
    {
	priv->gl.dp2_VertexBuffer = lpVertices;
	priv->gl.glFlags |= GLF_VERTEX_BUFFER;
    }
    else
    {
	priv->gl.dp2_lpVertices = lpVertices;
    }

    if (GL_CHECK_NEW_TNL(dwVertexType, priv->gl.dp2_dwVertexType))
	priv->gl.glFlags |= GLF_NEW_TNL_MODE;

    priv->gl.dp2_dwVertexType = dwVertexType;
    priv->gl.dp2_dwVertexOffset = dwVertexOffset;
    priv->gl.dp2_dwVertexLength = dwVertexLength;
    priv->gl.dp2_dwFlags = dwFlags;

    return D3D_OK;
}

static HRESULT WINAPI GL_start_dp2(LPVOID ctx, DWORD dwFlags,
				   DWORD dwVertexType, LPVOID lpVertices,
				   DWORD dwVertexOffset, DWORD dwVertexLength,
				   DWORD dwReqVertexBufSize, DWORD dwReqCommandBufSize)
{
    return GL_start_dp2_sd(ctx, dwFlags, dwVertexType, lpVertices,
			   dwVertexOffset, dwVertexLength,
			   dwReqVertexBufSize, dwReqCommandBufSize, NULL);
}

static HRESULT WINAPI GL_emit_dp2(LPVOID ctx, LPD3DHAL_DP2COMMAND cmd,
				  LPVOID data1, DWORD dwSize1, LPVOID data2, DWORD dwSize2)
{
    IDirect3DDeviceImpl* This = (IDirect3DDeviceImpl*)ctx;
    GL_D3D_PRIV_VAR(priv, This);
    GL_dp2_handler func;
    func = GL_dp2_func[cmd->bCommand].func;
    if (!func) {
	FIXME("handler for %s not implemented\n", GL_dp2_func[cmd->bCommand].name);
	/* return D3DERR_EXECUTE_FAILED; */
	return D3D_OK; /* pretend success */
    }
    if (priv->gl.block && GL_dp2_func[cmd->bCommand].type == STAT) {
	/* record to state block */
	TRACE("saving state %s\n", GL_dp2_func[cmd->bCommand].name);
	GL_record_cmd(This, &priv->gl.block->cmds, cmd, data1, dwSize1, data2, dwSize2);
	return D3D_OK;
    }
    if (!This->d3dp.in_scene) {
	if (func == D3D_GL_draw) return D3DERR_SCENE_NOT_IN_SCENE;
	/* record command */
	TRACE("recording command %s\n", GL_dp2_func[cmd->bCommand].name);
	GL_record_cmd(This, &priv->gl.cmds, cmd, data1, dwSize1, data2, dwSize2);
	return D3D_OK;
    }
    return func(This, cmd, data1, data2);
}

static HRESULT WINAPI GL_flush_dp2(LPVOID ctx)
{
    IDirect3DDeviceImpl* This = (IDirect3DDeviceImpl*)ctx;
    GL_D3D_PRIV_VAR(priv, This);

    if (!This->d3dp.in_scene) return D3D_OK;

    GL_replay_cmds(This, &priv->gl.cmds, REPLAY_EXEC|REPLAY_DEL);

    glFlush();
    CHECK_GL_ERROR("glFlush");
    return D3D_OK;
}

static HRESULT GL_replay_cmds(IDirect3DDeviceImpl *This, struct GL_D3D_saved_cmd **cmd, DWORD dwFlags)
{
    struct GL_D3D_saved_cmd *ncmd = *cmd;

    while (ncmd) {
	TRACE("replaying command %s\n", GL_dp2_func[ncmd->cmd.bCommand].name);
	if (dwFlags & REPLAY_EXEC) {
	    GL_dp2_func[ncmd->cmd.bCommand].func(This, &ncmd->cmd, ncmd->data1, ncmd->data2);
	    /* FIXME: what to do in case of an error? */
	}
	if (dwFlags & REPLAY_APPLY)
	    GL_apply_cmd(This, ncmd);

	ncmd = ncmd->next;
	if (dwFlags & REPLAY_DEL) GL_delete_cmd(This, cmd);
    }
    return D3D_OK;
}
